package gov.usgs.earthquake.nshmp.netcdf.www;

import java.nio.file.Path;
import java.util.List;

import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.gmm.NehrpSiteClass;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfDataFilesHazardCurves;
import gov.usgs.earthquake.nshmp.netcdf.www.HazardMetadata.HazardResponseMetadata;
import gov.usgs.earthquake.nshmp.netcdf.www.Request.RequestDataSiteClass;
import gov.usgs.earthquake.nshmp.netcdf.www.RequestHazardCurves.HazardRequestDataImt;
import gov.usgs.earthquake.nshmp.www.NshmpMicronautServlet;
import gov.usgs.earthquake.nshmp.www.ResponseBody;

import io.micronaut.context.annotation.Value;
import io.micronaut.context.event.StartupEvent;
import io.micronaut.core.annotation.Nullable;
import io.micronaut.http.HttpRequest;
import io.micronaut.http.HttpResponse;
import io.micronaut.http.MediaType;
import io.micronaut.http.annotation.Controller;
import io.micronaut.http.annotation.Get;
import io.micronaut.http.annotation.PathVariable;
import io.micronaut.http.annotation.QueryValue;
import io.micronaut.runtime.event.annotation.EventListener;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.media.Content;
import io.swagger.v3.oas.annotations.media.ExampleObject;
import io.swagger.v3.oas.annotations.media.Schema;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.tags.Tag;
import jakarta.inject.Inject;

/**
 * Micronaut controller for getting static hazards from a NetCDF file.
 *
 * @see NetcdfService
 *
 * @author U.S. Geological Survey
 */
@Tag(name = "Static Hazard Data")
@Controller("/hazard")
public class NetcdfController {

  @Inject
  private NshmpMicronautServlet servlet;

  /**
   * The path to a hazard NetCDF file
   */
  @Value("${nshmp-ws-static.netcdf-path}")
  Path netcdfPath;

  NetcdfServiceHazardCurves service;

  /**
   * Read in data type and return the appropriate service to use.
   */
  @EventListener
  void startup(StartupEvent event) {
    var netcdfDataFiles = new NetcdfDataFilesHazardCurves(netcdfPath);
    service = new NetcdfServiceHazardCurves(netcdfDataFiles);
  }

  /**
   * GET method to return static curves using slash delimited.
   *
   * @param request The HTTP request
   * @param longitude The longitude of the site
   * @param latitude Latitude of the site
   * @param siteClass The site class
   */
  @Operation(
      summary = "Returns static curves given a longitude, latitude, site class, and IMT",
      description = "### Returns hazard curves for a " +
          "user-specified latitude, longitude, site class, and IMT.\n" +

          "Enter the latitude and longitude, select the site class, " +
          "IMT, and format, and press `Execute`.\n" +

          "### Service Response Types\n" +
          "The web service can respond in JSON or CSV format.\n" +
          "<br><br>" +
          "Choose the format type by adding a `format` query to the service call:" +
          "`?format=JSON` or `?format=CSV`\n" +
          "<br><br>" +
          "Default: `JSON`\n\n" +

          "### Service call pattern\n" +
          "This service call is slashed delimited with pattern: " +
          "`/hazard/{longitude}/{latitude}/{siteClass}/{imt}`\n" +
          "<br><br>" +
          "Example: `/hazard/-118/34/BC/PGA`" +
          "<br><br>" +
          "CSV Example: `/hazard/-118/34/BC/PGA?format=CSV`",
      operationId = "hazard-by-imt")
  @ApiResponse(
      description = "Spatially interpolated hazard curves",
      responseCode = "200",
      content = {
          @Content(
              mediaType = MediaType.APPLICATION_JSON,
              schema = @Schema(implementation = ResponseByImt.class)),
          @Content(
              mediaType = MediaType.TEXT_CSV,
              examples = @ExampleObject(
                  name = "CSV",
                  value = "Static Hazard Curves\n" +
                      "\n" +
                      "Longitude,-104.99,Latitude,39.34,SiteClass,BC,Imt,PGA\n" +
                      "Ground Motion (g),0.00233,0.0035,0.00524,0.00786,0.0118, ...\n" +
                      "Annual Frequency of Exceedence,0.036386,0.026034,0.018125,0.012197, ...\n"))
      })
  @Get(uri = "/{longitude}/{latitude}/{siteClass}/{imt}{?format}",
      produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetSlashByImt(
      HttpRequest<?> request,
      @Schema(required = true) @PathVariable Double latitude,
      @Schema(required = true) @PathVariable Double longitude,
      @Schema(required = true) @PathVariable NehrpSiteClass siteClass,
      @Schema(required = true) @PathVariable Imt imt,
      @Schema(required = false, defaultValue = "JSON",
          implementation = ResponseFormat.class) @QueryValue @Nullable ResponseFormat format) {
    return doGet(request, longitude, latitude, siteClass, imt, format);
  }

  /**
   * GET method to return static curves using slash delimited.
   *
   * @param request The HTTP request
   * @param longitude The longitude of the site
   * @param latitude Latitude of the site
   * @param siteClass The site class
   */
  @Operation(
      summary = "Returns static curves given a longitude, latitude, and site class.",
      description = "### Returns hazard curves for a " +
          "user-specified latitude, longitude, and site class.\n" +

          "Enter the latitude and longitude, select the site class and format (optional), " +
          " and press `Execute`.\n" +

          "### Service Response Types\n" +
          "The web service can respond in JSON or CSV format.\n" +
          "<br><br>" +
          "Choose the format type by adding a `format` query to the service call:" +
          "`?format=JSON` or `?format=CSV`\n" +
          "<br><br>" +
          "Default: `JSON`\n\n" +

          "### Service call pattern\n" +
          "This service call is slashed delimited with pattern: " +
          "`/hazard/{longitude}/{latitude}/{siteClass}`\n" +
          "<br><br>" +
          "Example: `/hazard/-118/34/BC`" +
          "<br><br>" +
          "Example: `/hazard/-118/34/BC?format=CSV`",
      operationId = "hazard-by-siteclass")
  @ApiResponse(
      description = "Spatially interpolated hazard curves",
      responseCode = "200",
      content = {
          @Content(
              mediaType = MediaType.APPLICATION_JSON,
              schema = @Schema(implementation = ResponseBySiteClass.class)),
          @Content(
              mediaType = MediaType.TEXT_CSV,
              examples = @ExampleObject(
                  name = "CSV",
                  value = "Static Hazard Curves\n" +
                      "\n" +
                      "Longitude,-104.99,Latitude,39.34,SiteClass,BC,Imt,PGA\n" +
                      "Ground Motion (g),0.00233,0.0035,0.00524,0.00786,0.0118, ...\n" +
                      "Annual Frequency of Exceedence,0.036386,0.026034,0.018125,0.012197, ...\n" +
                      "\n" +
                      "..."))

      })
  @Get(uri = "/{longitude}/{latitude}/{siteClass}{?format}",
      produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetSlashBySite(
      HttpRequest<?> request,
      @Schema(required = true) @PathVariable Double latitude,
      @Schema(required = true) @PathVariable Double longitude,
      @Schema(required = true) @PathVariable NehrpSiteClass siteClass,
      @Schema(required = false, defaultValue = "JSON",
          implementation = ResponseFormat.class) @QueryValue @Nullable ResponseFormat format) {
    return doGet(request, longitude, latitude, siteClass, null, format);
  }

  /**
   * GET method to return hazard curves using slash delimited.
   *
   * @param request The HTTP request
   * @param longitude The longitude of the site
   * @param latitude Latitude of the site
   */
  @Operation(
      summary = "Returns static curves given a longitude and latitude",
      description = "### Returns hazard curves for all site class for a " +
          "user-specified latitude and longitude.\n" +

          "Enter the latitude and longitude, select the format (optional), and press `Execute`.\n" +

          "### Service Response Types\n" +
          "The web service can respond in JSON or CSV format.\n" +
          "<br><br>" +
          "Choose the format type by adding a `format` query to the service call:" +
          "`?format=JSON` or `?format=CSV`\n" +
          "<br><br>" +
          "Default: `JSON`\n\n" +

          "### Service call pattern\n" +
          "This service call is slashed delimited with pattern: " +
          "`/hazard/{longitude}/{latitude}`\n" +
          "<br><br>" +
          "Example: `/hazard/-118/34`" +
          "<br><br>" +
          "Example: `/hazard/-118/34?format=CSV`",
      operationId = "hazard")
  @ApiResponse(
      description = "Returns static curves from the NSHM NetCDF file",
      responseCode = "200",
      content = {
          @Content(
              mediaType = MediaType.APPLICATION_JSON,
              schema = @Schema(implementation = Response.class)),
          @Content(
              mediaType = MediaType.TEXT_CSV,
              examples = @ExampleObject(
                  name = "CSV",
                  value = "Static Hazard Curves\n" +
                      "\n" +
                      "Longitude,-104.99,Latitude,39.34,SiteClass,BC,Imt,PGA\n" +
                      "Ground Motion (g),0.00233,0.0035,0.00524,0.00786,0.0118, ...\n" +
                      "Annual Frequency of Exceedence,0.036386,0.026034,0.018125,0.012197, ...\n" +
                      "\n" +
                      "..."))

      })
  @Get(uri = "/{longitude}/{latitude}{?format}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetSlash(
      HttpRequest<?> request,
      @Schema(required = true) @PathVariable Double longitude,
      @Schema(required = true) @PathVariable Double latitude,
      @Schema(required = false, defaultValue = "JSON",
          implementation = ResponseFormat.class) @QueryValue @Nullable ResponseFormat format) {
    return doGet(request, longitude, latitude, null, null, format);
  }

  /**
   * GET method to return a static curve using URL query.
   *
   * @param request The HTTP request
   * @param nshm The NSHM to query
   * @param longitude The longitude of the site
   * @param latitude Latitude of the site
   * @param siteClass The site class (optional)
   */
  @Operation(
      summary = "Returns static curves given a longitude, latitude, site class, and IMT",
      description = "### Returns hazard curves for a " +
          "user-specified latitude, longitude, site class, and IMT.\n" +

          "Enter the latitude and longitude select the site class (optional) " +
          "IMT (optional), and format (optional), and press `Execute`.\n" +

          "### Service Response Types\n" +
          "The web service can respond in JSON or CSV format.\n" +
          "<br><br>" +
          "Choose the format type by adding a `format` query to the service call:" +
          "`?format=JSON` or `?format=CSV`\n" +
          "<br><br>" +
          "Default: `JSON`\n\n" +

          "### Service call pattern\n" +
          "This service call is query based with pattern: " +
          "`/hazard?longitude={number}&latitude={number}" +
          "&siteClass={string}&imt={string}&format={CSV|JSON}`\n" +
          "<br><br>" +
          "Example: `/hazard?longitude=-118&latitude=34&siteClass=A&imt=PGA`" +
          "<br><br>" +
          "CSV Example: `/hazard?longitude=-118&latitude=34&siteClass=A&imt=PGA?format=CSV`",
      operationId = "hazard-by-imt")
  @ApiResponse(
      description = "Spatially interpolated hazard curves",
      responseCode = "200",
      content = {
          @Content(
              mediaType = MediaType.APPLICATION_JSON,
              schema = @Schema(implementation = ResponseByImt.class)),
          @Content(
              mediaType = MediaType.TEXT_CSV,
              examples = @ExampleObject(
                  name = "CSV",
                  value = "Static Hazard Curves\n" +
                      "\n" +
                      "Longitude,-104.99,Latitude,39.34,SiteClass,BC,Imt,PGA\n" +
                      "Ground Motion (g),0.00233,0.0035,0.00524,0.00786,0.0118, ...\n" +
                      "Annual Frequency of Exceedence,0.036386,0.026034,0.018125,0.012197, ...\n" +
                      "\n" +
                      "..."))

      })
  @Get(uri = "{?longitude,latitude,siteClass,imt,format}",
      produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGet(
      HttpRequest<?> request,
      @Schema(required = true) @QueryValue @Nullable Double longitude,
      @Schema(required = true) @QueryValue @Nullable Double latitude,
      @QueryValue @Nullable NehrpSiteClass siteClass,
      @QueryValue @Nullable Imt imt,
      @Schema(required = false, defaultValue = "JSON",
          implementation = ResponseFormat.class) @QueryValue @Nullable ResponseFormat format) {
    var query = new HazardQuery(longitude, latitude, siteClass, imt, format);
    return service.handleServiceCall(request, query);
  }

  // For Swagger schema
  private static class ResponseByImt
      extends
      ResponseBody<HazardRequestDataImt, ResponseData<HazardResponseMetadata>> {}

  // For Swagger schema
  private static class ResponseBySiteClass
      extends
      ResponseBody<RequestDataSiteClass, List<ResponseData<HazardResponseMetadata>>> {}

  // For Swagger schema
  private static class Response
      extends
      ResponseBody<RequestDataSiteClass, List<List<ResponseData<HazardResponseMetadata>>>> {}
}
