package gov.usgs.earthquake.nshmp.netcdf.reader;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;

import java.io.IOException;
import java.nio.file.Path;
import java.util.Arrays;

import com.google.common.math.DoubleMath;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.LocationList;

import ucar.ma2.DataType;
import ucar.nc2.Attribute;
import ucar.nc2.Group;
import ucar.nc2.dataset.NetcdfDataset;
import ucar.nc2.dataset.NetcdfDatasets;

public class NetcdfUtils {

  /** TODO: Have this be dynamic from NetCDF file */
  public static final int LOCATION_PRECISION = 3;

  /** TODO: Have this be dynamic from NetCDF file */
  public static final int GROUND_MOTION_PRECISION = 3;

  /** TODO: Have this be dynamic from NetCDF file */
  public static final int HAZARD_PRECISION = 5;

  // Tolerance for longitude/latitude comparisons
  static final double LOCATION_TOLERANCE = 0.000001;

  /**
   * Creates a border going clockwise of the given longitudes and latitudes.
   *
   * @param longitudes The longitudes
   * @param latitudes The latitudes
   * @return
   */
  public static LocationList buildBorder(double[] longitudes, double[] latitudes) {
    var builder = LocationList.builder();

    for (var lat : latitudes) {
      builder.add(longitudes[0], lat);
    }

    // omit duplicate points at corners
    for (var i = 1; i < longitudes.length; i++) {
      builder.add(longitudes[i], latitudes[latitudes.length - 1]);
    }

    for (var i = latitudes.length - 2; i >= 0; i--) {
      builder.add(longitudes[longitudes.length - 1], latitudes[i]);
    }

    for (var i = longitudes.length - 2; i >= 0; i--) {
      builder.add(longitudes[i], latitudes[0]);
    }

    return builder.build();
  }

  /*
   * Calculate fractional distance from a1 to t, between a1 and a2
   */
  static double calcFrac(double a1, double a2, double t) {
    if (Math.abs(t - a1) < LOCATION_TOLERANCE) {
      // target value == a1
      return 0.0;
    } else if (Math.abs(t - a2) < LOCATION_TOLERANCE) {
      // target value == a2
      return 1.0;
    } else {
      // calculate fractional distance to t between a[i] and a[i+1]
      return (t - a1) / (a2 - a1);
    }
  }

  /*
   * Calculate fractional distance from a[i] to t, between a[i] and a[i+1]
   */
  static double calcGridFrac(double[] a, int i, double t) {
    return calcFrac(a[i], a[i + 1], t);
  }

  /**
   * Check that the X values are identical.
   *
   * @param a Sequence A
   * @param b Sequence B
   */
  public static void checkXySequence(XySequence a, XySequence b) {
    checkState(
        Arrays.equals(a.xValues().toArray(), b.xValues().toArray()),
        "Hazard curves xValues are not the same");
  }

  static XySequence cleanData(XySequence data, int precision) {
    var xs =
        data.xValues().map(x -> Double.isNaN(x) ? x : Maths.roundToDigits(x, precision)).toArray();
    var ys =
        data.yValues().map(y -> Double.isNaN(y) ? y : Maths.roundToDigits(y, precision)).toArray();
    return XySequence.create(xs, ys);
  }

  /**
   * Get a 1D array from a netCDF group.
   *
   * @param group The netCDF group
   * @param key The key to read from the group
   * @param dataType The data type to read
   * @throws IOException
   */
  static Object get1DArray(Group group, String key, DataType dataType) throws IOException {
    var var = group.findVariableLocal(key);
    checkNotNull(
        var,
        String.format("Could not find variable [%s] in group [%s]", key, group.getFullName()));

    return var.read().get1DJavaArray(dataType);
  }

  /**
   * Returns a {@code double[]} from a netCDF group
   *
   * @param group The netCDF group
   * @param key The key to read from the group
   * @throws IOException
   */
  static double[] getDoubleArray(Group group, String key) throws IOException {
    return (double[]) get1DArray(group, key, DataType.DOUBLE);
  }

  /*
   * find index of first element in a (sorted ascending) that is less than or
   * equal to target value t. If target value is equal to the maximum value in a
   * (the last element), an index of a.length - 1 is returned.
   */
  static int getIdxLTEQ(double[] a, double t) {
    // assumes array is in sorted order (a[i] < a[i+1])
    // make sure target is within the range of the array
    var n = a.length;

    if (t < a[0] || t > a[n - 1]) {
      // should never get here thanks to NshmpNetcdfCoordinates.checkCoords()
      throw new IllegalArgumentException(
          String.format("Target [%.4f] outside of valid range: %.4f <= t <= %.4f", t, a[0],
              a[n - 1]));
    }

    // if (t == a[0]) {
    if (DoubleMath.fuzzyEquals(a[0], t, LOCATION_TOLERANCE)) {
      return 0;
    }
    // if (t == a[n - 1]) {
    if (DoubleMath.fuzzyEquals(a[n - 1], t, LOCATION_TOLERANCE)) {
      return n - 2; // return second to last index number
    }

    var idx = Arrays.binarySearch(a, t);
    if (idx < 0) {
      // "exact" match not found, so use index of first value less than target
      // this is insertion_point - 1
      // returned idx = -insertion_point - 1
      idx = -1 * (idx + 1) - 1;
      // array[idx] <= target
    }

    return idx;
  }

  /**
   * Returns a {@code int[]} from a netCDF group
   *
   * @param group The netCDF group
   * @param key The key to read from the group
   * @throws IOException
   */
  static int[] getIntArray(Group group, String key) throws IOException {
    return (int[]) get1DArray(group, key, DataType.INT);
  }

  /**
   * Returns a {@code String[]} from a netCDF group
   *
   * @param group The netCDF group
   * @param key The key to read from the group
   * @throws IOException
   */
  static String[] getStringArray(Group group, String key) throws IOException {
    return (String[]) get1DArray(group, key, DataType.STRING);
  }

  public static Attribute readAttribute(String attributeKey, Path netcdfPath) {
    try (NetcdfDataset ncd = NetcdfDatasets.openDataset(netcdfPath.toString())) {
      Group group = ncd.getRootGroup();
      return group.attributes().findAttribute(attributeKey);
    } catch (IOException e) {
      throw new RuntimeException("Could not read Netcdf file [" + netcdfPath + " ]");
    }
  }

  public static class Key {
    public static final String DESCRIPTION = "description";
    public static final String GRID_STEP = "gridStep";
    public static final String ID = "id";
    public static final String LABEL = "label";
    public static final String DATA_TYPE = "dataType";
    public static final String GROUND_MOTION = "groundMotion";
    public static final String HAZARD = "hazard";
    public static final String HISTORY = "history";
    public static final String IMLS = "iml";
    public static final String IMT = "imt";
    public static final String LAT = "lat";
    public static final String LON = "lon";
    public static final String MAP = "map";
    public static final String REGION = "region";
    public static final String SCIENCE_BASE_INFO = "scienceBaseInfo";
    public static final String SITE_CLASS = "siteClass";
    public static final String SITES = "sites";
    public static final String VS30 = "vs30";
    public static final String YEAR = "year";
  }
}
